@@ -45,6 +45,7 @@
 #include "daemon.h"
 #include "engines.h"
 #include "stack.h"
+#include "sound.h"
 #include "notificationdaemon-dbus-glue.h"
 
 #define IMAGE_SIZE 48
@@ -82,13 +83,15 @@ typedef struct
 	gboolean paused;
 	guint id;
 	GtkWindow *nw;
-
+	Window src_window_xid;
 } NotifyTimeout;
 
 struct _NotifyDaemonPrivate
 {
 	guint next_id;
 	guint timeout_source;
+	GHashTable *idle_reposition_notify_ids;
+	GHashTable *monitored_window_hash;
 	GHashTable *notification_hash;
 	gboolean url_clicked_lock;
 	NotifyStack **stacks;
@@ -115,10 +118,19 @@ struct _DBusGMethodInvocation
 
 static void notify_daemon_finalize(GObject *object);
 static void _close_notification(NotifyDaemon *daemon, guint id,
-								gboolean hide_notification);
-static void _emit_closed_signal(GtkWindow *nw);
+								gboolean hide_notification,
+								NotifydClosedReason reason);
+static GdkFilterReturn _notify_x11_filter(GdkXEvent *xevent,
+										  GdkEvent *event,
+										  gpointer user_data);
+static void _emit_closed_signal(GtkWindow *nw, NotifydClosedReason reason);
 static void _action_invoked_cb(GtkWindow *nw, const char *key);
 static NotifyStackLocation get_stack_location_from_string(const char *slocation);
+static void sync_notification_position(NotifyDaemon *daemon, GtkWindow *nw,
+									   Window source);
+static void monitor_notification_source_windows(NotifyDaemon *daemon,
+												NotifyTimeout *nt,
+												Window source);
 
 G_DEFINE_TYPE(NotifyDaemon, notify_daemon, G_TYPE_OBJECT);
 
@@ -165,6 +177,10 @@ notify_daemon_init(NotifyDaemon *daemon)
 	daemon->priv->stacks_size = gdk_screen_get_n_monitors(screen);
 	daemon->priv->stacks = g_new0(NotifyStack *, daemon->priv->stacks_size);
 
+	daemon->priv->idle_reposition_notify_ids = g_hash_table_new(NULL, NULL);
+	daemon->priv->monitored_window_hash = g_hash_table_new(NULL, NULL);
+	gdk_window_add_filter(NULL, _notify_x11_filter, daemon);
+
 	for (i = 0; i < daemon->priv->stacks_size; i++)
 	{
 		daemon->priv->stacks[i] = notify_stack_new(daemon, screen,
@@ -182,6 +198,8 @@ notify_daemon_finalize(GObject *object)
 	NotifyDaemon *daemon       = NOTIFY_DAEMON(object);
 	GObjectClass *parent_class = G_OBJECT_CLASS(notify_daemon_parent_class);
 
+	g_hash_table_destroy(daemon->priv->monitored_window_hash);
+	g_hash_table_destroy(daemon->priv->idle_reposition_notify_ids);
 	g_hash_table_destroy(daemon->priv->notification_hash);
 	g_free(daemon->priv);
 
@@ -248,19 +266,23 @@ _action_invoked_cb(GtkWindow *nw, const char *key)
 	dbus_connection_send(dbus_conn, message, NULL);
 	dbus_message_unref(message);
 
-	_close_notification(daemon, id, TRUE);
+	_close_notification(daemon, id, TRUE, NOTIFYD_CLOSED_USER);
 }
 
 static void
-_emit_closed_signal(GtkWindow *nw)
+_emit_closed_signal(GtkWindow *nw, NotifydClosedReason reason)
 {
 	DBusMessage *message = create_signal(nw, "NotificationClosed");
+	dbus_message_append_args(message,
+							 DBUS_TYPE_UINT32, &reason,
+							 DBUS_TYPE_INVALID);
 	dbus_connection_send(dbus_conn, message, NULL);
 	dbus_message_unref(message);
 }
 
 static void
-_close_notification(NotifyDaemon *daemon, guint id, gboolean hide_notification)
+_close_notification(NotifyDaemon *daemon, guint id,
+					gboolean hide_notification, NotifydClosedReason reason)
 {
 	NotifyDaemonPrivate *priv = daemon->priv;
 	NotifyTimeout *nt;
@@ -269,7 +291,7 @@ _close_notification(NotifyDaemon *daemon, guint id, gboolean hide_notification)
 
 	if (nt != NULL)
 	{
-		_emit_closed_signal(nt->nw);
+		_emit_closed_signal(nt->nw, reason);
 
 		if (hide_notification)
 			theme_hide_notification(nt->nw);
@@ -281,7 +303,118 @@ _close_notification(NotifyDaemon *daemon, guint id, gboolean hide_notification)
 static void
 _notification_destroyed_cb(GtkWindow *nw, NotifyDaemon *daemon)
 {
-	_close_notification(daemon, NW_GET_NOTIFY_ID(nw), FALSE);
+	/*
+	 * This usually won't happen, but can if notification-daemon dies before
+	 * all notifications are closed. Mark them as expired.
+	 */
+	_close_notification(daemon, NW_GET_NOTIFY_ID(nw), FALSE,
+						NOTIFYD_CLOSED_EXPIRED);
+}
+
+typedef struct
+{
+	NotifyDaemon *daemon;
+	gint id;
+} IdleRepositionData;
+
+static gboolean
+idle_reposition_notification(gpointer datap)
+{
+	IdleRepositionData *data = (IdleRepositionData *)datap;
+	NotifyDaemon *daemon = data->daemon;
+	NotifyTimeout *nt;
+	gint notify_id;
+
+	notify_id = data->id;
+
+	/* Look up the timeout, if it's completed we don't need to do anything */
+	nt = (NotifyTimeout *)g_hash_table_lookup(daemon->priv->notification_hash,
+											  &notify_id);
+	if (nt != NULL) {
+		sync_notification_position(daemon, nt->nw, nt->src_window_xid);
+	}
+
+	g_hash_table_remove(daemon->priv->idle_reposition_notify_ids,
+						GINT_TO_POINTER(notify_id));
+	g_object_unref(daemon);
+	g_free(data);
+
+	return FALSE;
+}
+
+static void
+_queue_idle_reposition_notification(NotifyDaemon *daemon, gint notify_id)
+{
+	IdleRepositionData *data;
+	gpointer orig_key;
+	gpointer value;
+	guint idle_id;
+
+	/* Do we already have an idle update pending? */
+	if (g_hash_table_lookup_extended(daemon->priv->idle_reposition_notify_ids,
+									 GINT_TO_POINTER(notify_id),
+									 &orig_key, &value))
+	{
+		return;
+	}
+
+	data = g_new0(IdleRepositionData, 1);
+	data->daemon = g_object_ref(daemon);
+	data->id = notify_id;
+
+	/* We do this as a short timeout to avoid repositioning spam */
+	idle_id = g_timeout_add_full(G_PRIORITY_LOW, 50,
+								 idle_reposition_notification, data, NULL);
+	g_hash_table_insert(daemon->priv->idle_reposition_notify_ids,
+						GINT_TO_POINTER(notify_id), GUINT_TO_POINTER(idle_id));
+}
+
+static GdkFilterReturn
+_notify_x11_filter(GdkXEvent *xevent,
+				   GdkEvent *event,
+				   gpointer user_data)
+{
+	NotifyDaemon *daemon = NOTIFY_DAEMON(user_data);
+	XEvent *xev = (XEvent *)xevent;
+	gpointer orig_key;
+	gpointer value;
+	gint notify_id;
+	NotifyTimeout *nt;
+
+	if (xev->xany.type == DestroyNotify)
+	{
+		g_hash_table_remove(daemon->priv->monitored_window_hash,
+							GUINT_TO_POINTER(xev->xany.window));
+		return GDK_FILTER_CONTINUE;
+	}
+
+	if (!g_hash_table_lookup_extended(daemon->priv->monitored_window_hash,
+					 GUINT_TO_POINTER(xev->xany.window), &orig_key, &value))
+		return GDK_FILTER_CONTINUE;
+
+	notify_id = GPOINTER_TO_INT(value);
+
+	if (xev->xany.type == ConfigureNotify || xev->xany.type == MapNotify)
+	{
+		_queue_idle_reposition_notification(daemon, notify_id);
+	}
+	else if (xev->xany.type == ReparentNotify)
+	{
+		nt = (NotifyTimeout *)g_hash_table_lookup(
+			daemon->priv->notification_hash, &notify_id);
+
+		if (nt == NULL)
+			return GDK_FILTER_CONTINUE;
+
+		/*
+		 * If the window got reparented, we need to start monitoring the
+		 * new parents.
+		 */
+		monitor_notification_source_windows(daemon, nt, nt->src_window_xid);
+		sync_notification_position(daemon, nt->nw, nt->src_window_xid);
+	}
+
+	return GDK_FILTER_CONTINUE;
 }
 
 static void
@@ -349,7 +482,7 @@ _is_expired(gpointer key, gpointer value, gpointer data)
 	if (now_time > expiration_time)
 	{
 		theme_notification_tick(nt->nw, 0);
-		_emit_closed_signal(nt->nw);
+		_emit_closed_signal(nt->nw, NOTIFYD_CLOSED_EXPIRED);
 		return TRUE;
 	}
 	else if (nt->paused)
@@ -425,7 +558,7 @@ _calculate_timeout(NotifyDaemon *daemon, NotifyTimeout *nt, int timeout)
 	}
 }
 
-static guint
+static NotifyTimeout *
 _store_notification(NotifyDaemon *daemon, GtkWindow *nw, int timeout)
 {
 	NotifyDaemonPrivate *priv = daemon->priv;
@@ -455,7 +588,7 @@ _store_notification(NotifyDaemon *daemon, GtkWindow *nw, int timeout)
 	g_hash_table_insert(priv->notification_hash,
 						g_memdup(&id, sizeof(guint)), nt);
 
-	return id;
+	return nt;
 }
 
 static gboolean
@@ -623,7 +756,8 @@ _notify_daemon_process_icon_data(NotifyDaemon *daemon, GtkWindow *nw,
 	if (expected_len != tmp_array->len)
 	{
 		g_warning("_notify_daemon_process_icon_data expected image "
-				  "data to be of length %i but got a length of %i",
+				  "data to be of length %" G_GSIZE_FORMAT " but got a "
+				  "length of %u",
 				  expected_len, tmp_array->len);
 		return FALSE;
 	}
@@ -650,7 +784,8 @@ window_clicked_cb(GtkWindow *nw, GdkEventButton *button, NotifyDaemon *daemon)
 	}
 
 	_action_invoked_cb(nw, "default");
-	_close_notification(daemon, NW_GET_NOTIFY_ID(nw), TRUE);
+	_close_notification(daemon, NW_GET_NOTIFY_ID(nw), TRUE,
+						NOTIFYD_CLOSED_USER);
 }
 
 static void
@@ -772,19 +907,24 @@ static gboolean
 fullscreen_window_exists(GtkWidget *nw)
 {
 	WnckScreen *wnck_screen;
+	WnckWorkspace *wnck_workspace;
 	GList *l;
 
 	wnck_screen = wnck_screen_get(GDK_SCREEN_XNUMBER(
 		gdk_drawable_get_screen(GDK_DRAWABLE(GTK_WIDGET(nw)->window))));
 	wnck_screen_force_update(wnck_screen);
 
+	wnck_workspace = wnck_screen_get_active_workspace(wnck_screen);
+
 	for (l = wnck_screen_get_windows_stacked(wnck_screen);
 		 l != NULL;
 		 l = l->next)
 	{
 		WnckWindow *wnck_win = (WnckWindow *)l->data;
 
-		if (wnck_window_is_fullscreen(wnck_win))
+		if (wnck_window_is_on_workspace(wnck_win, wnck_workspace) &&
+			wnck_window_is_fullscreen(wnck_win) &&
+			wnck_window_is_active(wnck_win))
 		{
 			/*
 			 * Sanity check if the window is _really_ fullscreen to
@@ -805,6 +945,113 @@ fullscreen_window_exists(GtkWidget *nw)
 	return FALSE;
 }
 
+static Window
+get_window_parent(Display *display,
+				  Window window,
+				  Window *root)
+{
+	Window parent;
+	Window *children = NULL;
+	guint nchildren;
+	gboolean result;
+
+	gdk_error_trap_push();
+	result = XQueryTree(display, window, root, &parent, &children, &nchildren);
+	if (gdk_error_trap_pop() || !result)
+		return None;
+
+	if (children)
+		XFree(children);
+
+	return parent;
+}
+
+/*
+ * Recurse over X Window and parents, up to root, and start watching them
+ * for position changes.
+ */
+static void
+monitor_notification_source_windows(NotifyDaemon *daemon,
+									NotifyTimeout *nt,
+									Window source)
+{
+	Display *display = GDK_DISPLAY_XDISPLAY(gdk_display_get_default());
+	Window root = None;
+	Window parent;
+
+	/* Store the window in the timeout */
+	g_assert(nt != NULL);
+	nt->src_window_xid = source;
+
+	for (parent = get_window_parent(display, source, &root);
+		 parent != None && root != parent;
+		 parent = get_window_parent(display, parent, &root)) {
+
+		XSelectInput(display, parent, StructureNotifyMask);
+		g_hash_table_insert(daemon->priv->monitored_window_hash,
+							GUINT_TO_POINTER(parent), GINT_TO_POINTER(nt->id));
+	}
+}
+
+/* Use a source X Window ID to reposition a notification. */
+static void
+sync_notification_position(NotifyDaemon *daemon,
+						   GtkWindow *nw,
+						   Window source)
+{
+	Display *display = GDK_DISPLAY_XDISPLAY(gdk_display_get_default());
+	Status result;
+	Window root;
+	Window child;
+	int x, y;
+	unsigned int width, height;
+	unsigned int border_width, depth;
+
+	gdk_error_trap_push();
+
+	/* Get the root for this window */
+	result = XGetGeometry(display, source, &root, &x, &y,
+						  &width, &height, &border_width, &depth);
+	if (gdk_error_trap_pop() || !result)
+		return;
+
+	/*
+	 * Now calculate the offset coordinates for the source window from
+	 * the root.
+	 */
+	gdk_error_trap_push ();
+	result = XTranslateCoordinates(display, source, root, 0, 0,
+								   &x, &y, &child);
+	if (gdk_error_trap_pop() || !result)
+		return;
+
+	x += width  / 2;
+	y += height / 2;
+
+	theme_set_notification_arrow(nw, TRUE, x, y);
+	theme_move_notification(nw, x, y);
+	theme_show_notification(nw);
+
+	/*
+	 * We need to manually queue a draw here as the default theme recalculates
+	 * its position in the draw handler and moves the window (which seems
+	 * fairly broken), so just calling move/show above isn't enough to cause
+	 * its position to be calculated.
+	 */
+	gtk_widget_queue_draw(GTK_WIDGET(nw));
+}
+
+GQuark
+notify_daemon_error_quark(void)
+{
+	static GQuark q = 0;
+
+	if (q == 0)
+		q = g_quark_from_static_string("notification-daemon-error-quark");
+
+	return q;
+}
+
 gboolean
 notify_daemon_notify_handler(NotifyDaemon *daemon,
 							 const gchar *app_name,
@@ -824,8 +1071,11 @@ notify_daemon_notify_handler(NotifyDaemon *daemon,
 	gboolean new_notification = FALSE;
 	gint x = 0;
 	gint y = 0;
+	Window window_xid = None;
 	guint return_id;
 	gchar *sender;
+	gchar *sound_file = NULL;
+	gboolean sound_enabled;
 	gint i;
 
 	if (id > 0)
@@ -867,8 +1117,13 @@ notify_daemon_notify_handler(NotifyDaemon *daemon,
 	 *XXX This needs to handle file URIs and all that.
 	 */
 
+
+	if ((data = (GValue *)g_hash_table_lookup(hints, "window-xid")) != NULL)
+	{
+		window_xid = (Window)g_value_get_uint(data);
+	}
 	/* deal with x, and y hints */
-	if ((data = (GValue *)g_hash_table_lookup(hints, "x")) != NULL)
+	else if ((data = (GValue *)g_hash_table_lookup(hints, "x")) != NULL)
 	{
 		x = g_value_get_int(data);
 
@@ -879,6 +1134,63 @@ notify_daemon_notify_handler(NotifyDaemon *daemon,
 		}
 	}
 
+	/* Deal with sound hints */
+	sound_enabled = gconf_client_get_bool(gconf_client,
+										  GCONF_KEY_SOUND_ENABLED, NULL);
+	data = (GValue *)g_hash_table_lookup(hints, "suppress-sound");
+
+	if (data != NULL)
+	{
+		if (G_VALUE_HOLDS_BOOLEAN(data))
+			sound_enabled = !g_value_get_boolean(data);
+		else if (G_VALUE_HOLDS_INT(data))
+			sound_enabled = (g_value_get_int(data) != 0);
+		else
+		{
+			g_warning("suppress-sound is of type %s (expected bool or int)\n",
+					  g_type_name(G_VALUE_TYPE(data)));
+		}
+	}
+
+	if (sound_enabled)
+	{
+		data = (GValue *)g_hash_table_lookup(hints, "sound-file");
+
+		if (data != NULL)
+		{
+			sound_file = g_value_dup_string(data);
+
+			if (*sound_file == '\0' ||
+				!g_file_test(sound_file, G_FILE_TEST_EXISTS))
+			{
+				g_free(sound_file);
+				sound_file = NULL;
+			}
+		}
+
+		/*
+		 * TODO: If we don't have a sound_file yet, get the urgency hint, then
+		 *       get the corresponding system event sound
+		 *
+		 *       We will need to parse /etc/sound/events/gnome-2.soundlist
+		 *       and ~/.gnome2/sound/events/gnome-2.soundlist.
+		 */
+
+		/* If we don't have a sound file yet, use our gconf default */
+		if (sound_file == NULL)
+		{
+			sound_file = gconf_client_get_string(gconf_client,
+												 GCONF_KEY_DEFAULT_SOUND, NULL);
+			if (sound_file != NULL &&
+				(*sound_file == '\0' ||
+				 !g_file_test(sound_file, G_FILE_TEST_EXISTS)))
+			{
+				g_free(sound_file);
+				sound_file = NULL;
+			}
+		}
+	}
+
 	/* set up action buttons */
 	for (i = 0; actions[i] != NULL; i += 2)
 	{
@@ -957,7 +1269,15 @@ notify_daemon_notify_handler(NotifyDaemon *daemon,
 		}
 	}
 
-	if (use_pos_data)
+
+	if (window_xid != None)
+	{
+		/*
+		 * Do nothing here if we were passed an XID; we'll call
+		 * sync_notification_position later.
+		 */
+	}
+	else if (use_pos_data)
 	{
 		/*
 		 * Typically, the theme engine will set its own position based on
@@ -983,13 +1303,34 @@ notify_daemon_notify_handler(NotifyDaemon *daemon,
 		notify_stack_add_window(priv->stacks[monitor], nw, new_notification);
 	}
 
+	if (id == 0)
+	{
+		nt = _store_notification(daemon, nw, timeout);
+		return_id = nt->id;
+	}
+	else
+		return_id = id;
+
+	/*
+	 * If we have a source Window XID, start monitoring the tree
+	 * for changes, and reposition the window based on the source
+	 * window.  We need to do this after return_id is calculated.
+	 */
+	if (window_xid != None)
+	{
+		monitor_notification_source_windows(daemon, nt, window_xid);
+		sync_notification_position(daemon, nw, window_xid);
+	}
+
 	if (!screensaver_active(GTK_WIDGET(nw)) &&
 		!fullscreen_window_exists(GTK_WIDGET(nw)))
 	{
 		theme_show_notification(nw);
+		if (sound_file != NULL)
+			sound_play(sound_file);
 	}
 
-	return_id = (id == 0 ? _store_notification(daemon, nw, timeout) : id);
+	g_free(sound_file);
 
 #if CHECK_DBUS_VERSION(0, 60)
 	sender = dbus_g_method_get_sender(context);
@@ -1015,9 +1356,15 @@ gboolean
 notify_daemon_close_notification_handler(NotifyDaemon *daemon,
 										 guint id, GError **error)
 {
-	_close_notification(daemon, id, TRUE);
-
-	return TRUE;
+	if (id == 0)
+	{
+		g_set_error(error, notify_daemon_error_quark(), 100,
+					_("%u is not a valid notification ID"), id);
+		return FALSE;
+	} else {
+		_close_notification(daemon, id, TRUE, NOTIFYD_CLOSED_API);
+		return TRUE;
+	}
 }
 
 gboolean
@@ -1045,7 +1392,7 @@ notify_daemon_get_server_information(NotifyDaemon *daemon,
 	*out_name     = g_strdup("Notification Daemon");
 	*out_vendor   = g_strdup("Galago Project");
 	*out_version  = g_strdup(VERSION);
-	*out_spec_ver = g_strdup("0.9");
+	*out_spec_ver = g_strdup("1.0");
 
 	return TRUE;
 }
@@ -1067,6 +1414,8 @@ main(int argc, char **argv)
 
 	g_log_set_always_fatal(G_LOG_LEVEL_ERROR | G_LOG_LEVEL_CRITICAL);
 
+	sound_init();
+
 	gtk_init(&argc, &argv);
 	gconf_init(argc, argv, NULL);
 
