@@ -16,6 +16,7 @@ typedef struct
 	GtkWidget *icon;
 	GtkWidget *content_hbox;
 	GtkWidget *summary_label;
+	GtkWidget *close_button;
 	GtkWidget *body_label;
 	GtkWidget *actions_box;
 	GtkWidget *last_sep;
@@ -80,16 +81,11 @@ enum
 #define BACKGROUND_OPACITY    0.92
 #define BOTTOM_GRADIENT_HEIGHT 30
 
-#if GTK_CHECK_VERSION(2, 8, 0)
-# define USE_CAIRO
-#endif
-
 #if GTK_CHECK_VERSION(2, 10, 0)
 # define USE_COMPOSITE
 #endif
 
 
-#ifdef USE_CAIRO
 static void
 fill_background(GtkWidget *widget, WindowData *windata, cairo_t *cr)
 {
@@ -131,21 +127,7 @@ fill_background(GtkWidget *widget, WindowData *windata, cairo_t *cr)
 	cairo_pattern_destroy(gradient);
 #endif
 }
-#else /* !USE_CAIRO */
-static void
-fill_background(GtkWidget *widget, WindowData *windata)
-{
-	GtkStyle *style = gtk_widget_get_style(windata->win);
 
-	gdk_draw_rectangle(GDK_DRAWABLE(widget->window),
-					   style->base_gc[GTK_STATE_NORMAL], TRUE,
-					   0, 0,
-					   widget->allocation.width,
-					   widget->allocation.height);
-}
-#endif
-
-#ifdef USE_CAIRO
 static void
 draw_stripe(GtkWidget *widget, WindowData *windata, cairo_t *cr)
 {
@@ -193,57 +175,22 @@ draw_stripe(GtkWidget *widget, WindowData *windata, cairo_t *cr)
 	cairo_fill(cr);
 #endif
 }
-#else /* !USE_CAIRO */
-static void
-draw_stripe(GtkWidget *widget, WindowData *windata)
-{
-	GtkStyle *style = gtk_widget_get_style(widget);
-	gboolean custom_gc = FALSE;
-	GdkColor color;
-	GdkGC *gc;
-
-	switch (windata->urgency)
-	{
-		case URGENCY_LOW: // LOW
-			gc = style->bg_gc[GTK_STATE_NORMAL];
-			break;
-
-		case URGENCY_CRITICAL: // CRITICAL
-			custom_gc = TRUE;
-			gc = gdk_gc_new(GDK_DRAWABLE(widget->window));
-			gdk_color_parse("#CC0000", &color);
-			gdk_gc_set_rgb_fg_color(gc, &color);
-			break;
-
-		case URGENCY_NORMAL: // NORMAL
-		default:
-			gc = style->bg_gc[GTK_STATE_SELECTED];
-			break;
-	}
-
-
-	gdk_draw_rectangle(widget->window, gc, TRUE,
-					   windata->main_hbox->allocation.x + 1,
-					   windata->main_hbox->allocation.y + 1,
-					   STRIPE_WIDTH,
-					   windata->main_hbox->allocation.height - 2);
-
-	if (custom_gc)
-		g_object_unref(G_OBJECT(gc));
-}
-#endif
 
 static GtkArrowType
 get_notification_arrow_type(GtkWidget *nw)
 {
 	WindowData *windata = g_object_get_data(G_OBJECT(nw), "windata");
-	int screen_height;
+	GdkScreen *screen;
+	GdkRectangle monitor_geometry;
+	int monitor;
 
-	screen_height = gdk_screen_get_height(
-		gdk_drawable_get_screen(GDK_DRAWABLE(nw->window)));
+	screen = gdk_drawable_get_screen(GDK_DRAWABLE(nw->window));
+	monitor = gdk_screen_get_monitor_at_point(screen, windata->point_x,
+											  windata->point_y);
+	gdk_screen_get_monitor_geometry(screen, monitor, &monitor_geometry);
 
-	if (windata->point_y + windata->height + DEFAULT_ARROW_HEIGHT >
-		screen_height)
+	if (windata->point_y - monitor_geometry.y + windata->height +
+		DEFAULT_ARROW_HEIGHT > monitor_geometry.height)
 	{
 		return GTK_ARROW_DOWN;
 	}
@@ -268,63 +215,68 @@ create_border_with_arrow(GtkWidget *nw, WindowData *windata)
 	int width;
 	int height;
 	int y;
+	int norm_point_x;
+	int norm_point_y;
 	GtkArrowType arrow_type;
 	GdkScreen *screen;
-	int screen_width;
-	int screen_height;
 	int arrow_side1_width = DEFAULT_ARROW_WIDTH / 2;
 	int arrow_side2_width = DEFAULT_ARROW_WIDTH / 2;
 	int arrow_offset = DEFAULT_ARROW_OFFSET;
 	GdkPoint *shape_points = NULL;
 	int i = 0;
+	int monitor;
+	GdkRectangle monitor_geometry;
 
 	width  = windata->width;
 	height = windata->height;
 
-	screen        = gdk_drawable_get_screen(GDK_DRAWABLE(nw->window));
-	screen_width  = gdk_screen_get_width(screen);
-	screen_height = gdk_screen_get_height(screen);
+	screen  = gdk_drawable_get_screen(GDK_DRAWABLE(nw->window));
+	monitor = gdk_screen_get_monitor_at_point(screen,
+											  windata->point_x,
+											  windata->point_y);
+	gdk_screen_get_monitor_geometry(screen, monitor, &monitor_geometry);
 
 	windata->num_border_points = 5;
 
 	arrow_type = get_notification_arrow_type(windata->win);
 
+	norm_point_x = windata->point_x - monitor_geometry.x;
+	norm_point_y = windata->point_y - monitor_geometry.y;
+
 	/* Handle the offset and such */
 	switch (arrow_type)
 	{
 		case GTK_ARROW_UP:
 		case GTK_ARROW_DOWN:
-			if (windata->point_x < arrow_side1_width)
+			if (norm_point_x < arrow_side1_width)
 			{
 				arrow_side1_width = 0;
 				arrow_offset = 0;
 			}
-			else if (windata->point_x > screen_width - arrow_side2_width)
+			else if (norm_point_x > monitor_geometry.width - arrow_side2_width)
 			{
 				arrow_side2_width = 0;
 				arrow_offset = width - arrow_side1_width;
 			}
 			else
 			{
-				if (windata->point_x - arrow_side2_width + width >=
-					screen_width)
+				if (norm_point_x - arrow_side2_width + width >=
+					monitor_geometry.width)
 				{
-					arrow_offset =
-						width - arrow_side1_width - arrow_side2_width -
-						(screen_width - MAX(windata->point_x +
-											arrow_side1_width,
-											screen_width -
-											DEFAULT_ARROW_OFFSET));
+					arrow_offset = width - monitor_geometry.width +
+						           norm_point_x;
 				}
 				else
 				{
-					arrow_offset = MIN(windata->point_x - arrow_side1_width,
+					arrow_offset = MIN(norm_point_x - arrow_side1_width,
 									   DEFAULT_ARROW_OFFSET);
 				}
 
 				if (arrow_offset == 0 ||
 					arrow_offset == width - arrow_side1_width)
+				{
 					windata->num_border_points++;
+				}
 				else
 					windata->num_border_points += 2;
 			}
@@ -348,7 +300,6 @@ create_border_with_arrow(GtkWidget *nw, WindowData *windata)
 			windata->drawn_arrow_middle_x = arrow_offset + arrow_side1_width;
 			windata->drawn_arrow_end_x = arrow_offset + arrow_side1_width +
 										 arrow_side2_width;
-
 			if (arrow_type == GTK_ARROW_UP)
 			{
 				windata->drawn_arrow_begin_y = DEFAULT_ARROW_HEIGHT;
@@ -439,15 +390,15 @@ create_border_with_arrow(GtkWidget *nw, WindowData *windata)
 
 		case GTK_ARROW_LEFT:
 		case GTK_ARROW_RIGHT:
-			if (windata->point_y < arrow_side1_width)
+			if (norm_point_y < arrow_side1_width)
 			{
 				arrow_side1_width = 0;
-				arrow_offset = windata->point_y;
+				arrow_offset = norm_point_y;
 			}
-			else if (windata->point_y > screen_height - arrow_side2_width)
+			else if (norm_point_y > monitor_geometry.height - arrow_side2_width)
 			{
 				arrow_side2_width = 0;
-				arrow_offset = windata->point_y - arrow_side1_width;
+				arrow_offset = norm_point_y - arrow_side1_width;
 			}
 			break;
 
@@ -463,7 +414,6 @@ create_border_with_arrow(GtkWidget *nw, WindowData *windata)
 	g_free(shape_points);
 }
 
-#ifdef USE_CAIRO
 static void
 draw_border(GtkWidget *widget, WindowData *windata, cairo_t *cr)
 {
@@ -501,56 +451,21 @@ draw_border(GtkWidget *widget, WindowData *windata, cairo_t *cr)
 
 	cairo_stroke(cr);
 }
-#else /* !USE_CAIRO */
-static void
-draw_border(GtkWidget *widget,
-			WindowData *windata)
-{
-	if (windata->gc == NULL)
-	{
-		GdkColor color;
-
-		windata->gc = gdk_gc_new(widget->window);
-		gdk_color_parse("black", &color);
-		gdk_gc_set_rgb_fg_color(windata->gc, &color);
-	}
-
-	if (windata->has_arrow)
-	{
-		create_border_with_arrow(windata->win, windata);
-
-		gdk_draw_polygon(widget->window, windata->gc, FALSE,
-						 windata->border_points, windata->num_border_points);
-		gdk_window_shape_combine_region(windata->win->window,
-										windata->window_region,
-										0, 0);
-		g_free(windata->border_points);
-		windata->border_points = NULL;
-	}
-	else
-	{
-		gdk_draw_rectangle(widget->window, windata->gc, FALSE,
-						   0, 0, windata->width - 1, windata->height - 1);
-	}
-
-}
-#endif
 
 static gboolean
 paint_window(GtkWidget *widget,
 			 GdkEventExpose *event,
 			 WindowData *windata)
 {
+	cairo_t *context;
+	cairo_surface_t *surface;
+	cairo_t *cr;
+
 	if (windata->width == 0) {
 		windata->width = windata->win->allocation.width;
 		windata->height = windata->win->allocation.height;
 	}
 
-#ifdef USE_CAIRO
-	cairo_t *context;
-	cairo_surface_t *surface;
-	cairo_t *cr;
-
 	context = gdk_cairo_create(widget->window);
 
 	cairo_set_operator(context, CAIRO_OPERATOR_SOURCE);
@@ -569,11 +484,6 @@ paint_window(GtkWidget *widget,
 	cairo_paint(context);
 	cairo_surface_destroy(surface);
 	cairo_destroy(context);
-#else /* !USE_CAIRO */
-	fill_background(widget, windata);
-	draw_border(widget, windata);
-	draw_stripe(widget, windata);
-#endif
 
 	return FALSE;
 }
@@ -666,6 +576,7 @@ create_notification(UrlClickedCb url_clicked)
 	GtkWidget *image;
 	GtkWidget *alignment;
 	AtkObject *atkobj;
+	GtkRcStyle *rcstyle;
 	WindowData *windata;
 #ifdef USE_COMPOSITE
 	GdkColormap *colormap;
@@ -692,6 +603,8 @@ create_notification(UrlClickedCb url_clicked)
 #endif
 
 	gtk_window_set_title(GTK_WINDOW(win), "Notification");
+	gtk_window_set_type_hint(GTK_WINDOW(win),
+							 GDK_WINDOW_TYPE_HINT_NOTIFICATION);
 	gtk_widget_add_events(win, GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK);
 	gtk_widget_realize(win);
 	gtk_widget_set_size_request(win, WIDTH, -1);
@@ -762,15 +675,25 @@ create_notification(UrlClickedCb url_clicked)
 	atk_object_set_description(atkobj, "Notification summary text.");
 
 	/* Add the close button */
+	alignment = gtk_alignment_new(1, 0, 0, 0);
+	gtk_widget_show(alignment);
+	gtk_box_pack_start(GTK_BOX(hbox), alignment, FALSE, FALSE, 0);
+
 	close_button = gtk_button_new();
+	windata->close_button = close_button;
 	gtk_widget_show(close_button);
-	gtk_box_pack_start(GTK_BOX(hbox), close_button, FALSE, FALSE, 0);
+	gtk_container_add(GTK_CONTAINER(alignment), close_button);
 	gtk_button_set_relief(GTK_BUTTON(close_button), GTK_RELIEF_NONE);
 	gtk_container_set_border_width(GTK_CONTAINER(close_button), 0);
-	gtk_widget_set_size_request(close_button, 20, 20);
+	//gtk_widget_set_size_request(close_button, 20, 20);
 	g_signal_connect_swapped(G_OBJECT(close_button), "clicked",
 							 G_CALLBACK(gtk_widget_destroy), win);
 
+	rcstyle = gtk_rc_style_new();
+	rcstyle->xthickness = rcstyle->ythickness = 0;
+	gtk_widget_modify_style(close_button, rcstyle);
+	gtk_rc_style_unref(rcstyle);
+
 	atkobj = gtk_widget_get_accessible(close_button);
 	atk_action_set_description(ATK_ACTION(atkobj), 0,
 							   "Closes the notification.");
@@ -829,7 +752,7 @@ set_notification_hints(GtkWindow *nw, GHashTable *hints)
 
 	value = (GValue *)g_hash_table_lookup(hints, "urgency");
 
-	if (value != NULL)
+	if (value != NULL && G_VALUE_HOLDS_UCHAR(value))
 	{
 		windata->urgency = g_value_get_uchar(value);
 
@@ -866,11 +789,15 @@ notification_tick(GtkWindow *nw, glong remaining)
 void
 set_notification_text(GtkWindow *nw, const char *summary, const char *body)
 {
-	char *str;
+	char *str, *quoted;
+	GtkRequisition req;
 	WindowData *windata = g_object_get_data(G_OBJECT(nw), "windata");
 	g_assert(windata != NULL);
 
-	str = g_strdup_printf("<b><big>%s</big></b>", summary);
+	quoted = g_markup_escape_text(summary, -1);
+	str = g_strdup_printf("<b><big>%s</big></b>", quoted);
+	g_free(quoted);
+
 	gtk_label_set_markup(GTK_LABEL(windata->summary_label), str);
 	g_free(str);
 
@@ -883,10 +810,24 @@ set_notification_text(GtkWindow *nw, const char *summary, const char *body)
 
 	update_content_hbox_visibility(windata);
 
+	if (body != NULL && *body != '\0')
+	{
+		gtk_widget_size_request (windata->iconbox, &req);
+		gtk_widget_set_size_request(
+			windata->body_label,
+			/* -1: border width for
+			   -6: spacing for hbox */
+			WIDTH - (1*2) - (10*2) - req.width - 6,
+			-1);
+	}
+
+	gtk_widget_size_request (windata->close_button, &req);
 	gtk_widget_set_size_request(
-		((body != NULL && *body == '\0')
-		 ? windata->body_label : windata->summary_label),
-		WIDTH - (IMAGE_SIZE + IMAGE_PADDING) - 10,
+		windata->summary_label,
+		/* -1: main_vbox border width
+		   -10: vbox border width
+		   -6: spacing for hbox */
+		WIDTH - (1*2) - (10*2) - SPACER_LEFT - req.width - (6*2),
 		-1);
 }
 
@@ -933,11 +874,10 @@ countdown_expose_cb(GtkWidget *pie, GdkEventExpose *event,
 					WindowData *windata)
 {
 	GtkStyle *style = gtk_widget_get_style(windata->win);
-
-#ifdef USE_CAIRO
 	cairo_t *context;
 	cairo_surface_t *surface;
 	cairo_t *cr;
+
 	context = gdk_cairo_create(GDK_DRAWABLE(windata->pie_countdown->window));
 	cairo_set_operator(context, CAIRO_OPERATOR_SOURCE);
 	surface = cairo_surface_create_similar(
@@ -948,15 +888,11 @@ countdown_expose_cb(GtkWidget *pie, GdkEventExpose *event,
 	cr = cairo_create(surface);
 
 	fill_background(pie, windata, cr);
-#else /* !USE_CAIRO */
-	fill_background(pie, windata);
-#endif
 
 	if (windata->timeout > 0)
 	{
 		gdouble pct = (gdouble)windata->remaining / (gdouble)windata->timeout;
 
-#ifdef USE_CAIRO
 		gdk_cairo_set_source_color(cr, &style->bg[GTK_STATE_ACTIVE]);
 
 		cairo_move_to(cr, PIE_RADIUS, PIE_RADIUS);
@@ -964,21 +900,13 @@ countdown_expose_cb(GtkWidget *pie, GdkEventExpose *event,
 						   -G_PI_2, -(pct * G_PI * 2) - G_PI_2);
 		cairo_line_to(cr, PIE_RADIUS, PIE_RADIUS);
 		cairo_fill(cr);
-#else /* !USE_CAIRO */
-		gdk_draw_arc(GDK_DRAWABLE(windata->pie_countdown->window),
-					 style->bg_gc[GTK_STATE_ACTIVE], TRUE,
-					 0, 0, PIE_WIDTH, PIE_HEIGHT,
-					 90 * 64, pct * 360.0 * 64.0);
-#endif
 	}
 
-#ifdef USE_CAIRO
 	cairo_destroy(cr);
 	cairo_set_source_surface(context, surface, 0, 0);
 	cairo_paint(context);
 	cairo_surface_destroy(surface);
 	cairo_destroy(context);
-#endif
 
 	return TRUE;
 }
